@@ -19,26 +19,15 @@
 
 package org.apache.james.jdkim;
 
-import java.io.IOException;
-import java.io.InputStream;
-import java.security.InvalidKeyException;
-import java.security.NoSuchAlgorithmException;
-import java.security.PublicKey;
-import java.security.Signature;
-import java.security.SignatureException;
-import java.util.Arrays;
-import java.util.HashMap;
-import java.util.Hashtable;
-import java.util.Iterator;
-import java.util.LinkedList;
-import java.util.List;
-import java.util.Map;
+import static org.apache.james.jdkim.DKIMCommon.signatureCheck;
 
 import org.apache.james.jdkim.api.BodyHasher;
 import org.apache.james.jdkim.api.Headers;
 import org.apache.james.jdkim.api.PublicKeyRecord;
 import org.apache.james.jdkim.api.PublicKeyRecordRetriever;
+import org.apache.james.jdkim.api.Result;
 import org.apache.james.jdkim.api.SignatureRecord;
+import org.apache.james.jdkim.exceptions.CompositeFailException;
 import org.apache.james.jdkim.exceptions.FailException;
 import org.apache.james.jdkim.exceptions.PermFailException;
 import org.apache.james.jdkim.exceptions.TempFailException;
@@ -49,10 +38,28 @@ import org.apache.james.jdkim.impl.Message;
 import org.apache.james.jdkim.impl.MultiplexingPublicKeyRecordRetriever;
 import org.apache.james.jdkim.tagvalue.PublicKeyRecordImpl;
 import org.apache.james.jdkim.tagvalue.SignatureRecordImpl;
+import org.apache.james.jdkim.tagvalue.SignatureRecordTemplate;
 
-public class DKIMVerifier extends DKIMCommon {
+import java.io.IOException;
+import java.io.InputStream;
+import java.security.InvalidKeyException;
+import java.security.NoSuchAlgorithmException;
+import java.security.PublicKey;
+import java.security.Signature;
+import java.security.SignatureException;
+import java.util.ArrayList;
+import java.util.Arrays;
+import java.util.HashMap;
+import java.util.Hashtable;
+import java.util.Iterator;
+import java.util.LinkedList;
+import java.util.List;
+import java.util.Map;
 
-    private PublicKeyRecordRetriever publicKeyRecordRetriever;
+public class DKIMVerifier {
+
+    private final PublicKeyRecordRetriever publicKeyRecordRetriever;
+    private final List<Result> result = new ArrayList<>();
 
     public DKIMVerifier() {
         this.publicKeyRecordRetriever = new MultiplexingPublicKeyRecordRetriever(
@@ -87,8 +94,7 @@ public class DKIMVerifier extends DKIMCommon {
         if (records == null || records.isEmpty()) {
             lastError = "no key for signature";
         } else {
-            for (Iterator<String> i = records.iterator(); i.hasNext();) {
-                String record = i.next();
+            for (String record : records) {
                 try {
                     PublicKeyRecord pk = newPublicKeyRecord(record);
                     pk.validate();
@@ -108,9 +114,9 @@ public class DKIMVerifier extends DKIMCommon {
 
     /**
      * asserts applicability of a signature record the a public key record.
-     * throws an 
-     * 
-     * @param pkr public key record
+     * throws an
+     *
+     * @param pkr  public key record
      * @param sign signature record
      * @throws PermFailException when the keys are not applicable
      */
@@ -120,47 +126,45 @@ public class DKIMVerifier extends DKIMCommon {
                     .matches()) {
                 throw new PermFailException("inapplicable key identity local="
                         + sign.getIdentityLocalPart() + " Pattern: "
-                        + pkr.getGranularityPattern().pattern(), sign.getIdentity().toString());
+                        + pkr.getGranularityPattern().pattern(), sign);
             }
-    
+
             if (!pkr.isHashMethodSupported(sign.getHashMethod())) {
                 throw new PermFailException("inappropriate hash for a="
-                        + sign.getHashKeyType() + "/" + sign.getHashMethod(), sign.getIdentity().toString());
+                        + sign.getHashKeyType() + "/" + sign.getHashMethod(), sign);
             }
             if (!pkr.isKeyTypeSupported(sign.getHashKeyType())) {
                 throw new PermFailException("inappropriate key type for a="
-                        + sign.getHashKeyType() + "/" + sign.getHashMethod(), sign.getIdentity().toString());
+                        + sign.getHashKeyType() + "/" + sign.getHashMethod(), sign);
             }
-    
+
             if (pkr.isDenySubdomains()) {
                 if (!sign.getIdentity().toString().toLowerCase().endsWith(
                         ("@" + sign.getDToken()).toLowerCase())) {
                     throw new PermFailException(
-                            "AUID in subdomain of SDID is not allowed by the public key record.", sign.getIdentity().toString());
+                            "AUID in subdomain of SDID is not allowed by the public key record.", sign);
                 }
             }
         } catch (IllegalStateException e) {
-            throw new PermFailException("Invalid public key: "+e.getMessage(), sign.getIdentity().toString());
+            throw new PermFailException("Invalid public key: " + e.getMessage(), sign);
         }
     }
 
     /**
      * Iterates through signature's declared lookup method
-     * 
-     * @param sign
-     *                the signature record
+     *
+     * @param sign the signature record
      * @return an "applicable" PublicKeyRecord
-     * @throws TempFailException
-     * @throws PermFailException
+     * @throws TempFailException For temporary error
+     * @throws PermFailException For unrecoverable failure
      */
     public PublicKeyRecord publicRecordLookup(SignatureRecord sign)
             throws TempFailException, PermFailException {
-        // System.out.println(sign);
         PublicKeyRecord key = null;
         TempFailException lastTempFailure = null;
         PermFailException lastPermFailure = null;
         for (Iterator<CharSequence> rlm = sign.getRecordLookupMethods().iterator(); key == null
-                && rlm.hasNext();) {
+                && rlm.hasNext(); ) {
             CharSequence method = rlm.next();
             try {
                 PublicKeyRecordRetriever pkrr = getPublicKeyRecordRetriever();
@@ -181,16 +185,16 @@ public class DKIMVerifier extends DKIMCommon {
         }
         if (key == null) {
             if (lastTempFailure != null) {
-                if (sign != null) lastTempFailure.setRelatedRecordIdentity(sign.getIdentity().toString());
+                lastTempFailure.setRelatedRecord(sign);
                 throw lastTempFailure;
             } else if (lastPermFailure != null) {
-                if (sign != null) lastPermFailure.setRelatedRecordIdentity(sign.getIdentity().toString());
+                lastPermFailure.setRelatedRecord(sign);
                 throw lastPermFailure;
             }            // this is unexpected because the publicKeySelector always returns
             // null or exception
             else {
                 throw new PermFailException(
-                        "no key for signature [unexpected condition]", sign.getIdentity().toString());
+                        "no key for signature [unexpected condition]", sign);
             }
         }
         return key;
@@ -199,40 +203,36 @@ public class DKIMVerifier extends DKIMCommon {
     /**
      * Verifies all of the DKIM-Signature records declared in the supplied input
      * stream
-     * 
-     * @param is
-     *                inputStream
+     *
+     * @param is inputStream
      * @return a list of verified signature records.
-     * @throws IOException
-     * @throws FailException
-     *                 if no signature can be verified
+     * @throws IOException If error occurs handling data
+     * @throws FailException if no signature can be verified
      */
     public List<SignatureRecord> verify(InputStream is) throws IOException,
             FailException {
         Message message;
         try {
-	        try {
-	            message = new Message(is);
-            } catch (RuntimeException e) {
-            	throw e;
-            } catch (IOException e) {
-            	throw e;
+            try {
+                message = new Message(is);
+            } catch (RuntimeException | IOException e) {
+                throw e;
             } catch (Exception e1) {
-            	// This can only be a MimeException but we don't declare to allow usage of
-            	// DKIMSigner without Mime4J dependency.
-	            throw new PermFailException("Mime parsing exception: "
-	                    + e1.getMessage(), e1);
-	        }
-	        try {
-	            return verify(message, message.getBodyInputStream());
-	        } finally {
-	            message.dispose();
-	        }
+                // This can only be a MimeException but we don't declare to allow usage of
+                // DKIMSigner without Mime4J dependency.
+                throw new PermFailException("Mime parsing exception: "
+                        + e1.getMessage(), e1);
+            }
+            try {
+                return verify(message, message.getBodyInputStream());
+            } finally {
+                message.dispose();
+            }
         } finally {
             is.close();
         }
     }
-    
+
     public BodyHasher newBodyHasher(Headers messageHeaders) throws FailException {
         List<String> fields = messageHeaders.getFields("DKIM-Signature");
         if (fields == null || fields.isEmpty()) {
@@ -242,46 +242,42 @@ public class DKIMVerifier extends DKIMCommon {
         // For each DKIM-signature we prepare an hashjob.
         // We calculate all hashes concurrently so to read
         // the inputstream only once.
-        Map<String, BodyHasherImpl> bodyHashJobs = new HashMap<String, BodyHasherImpl>();
-        Hashtable<String, FailException> signatureExceptions = new Hashtable<String, FailException>();
-        for (Iterator<String> i = fields.iterator(); i.hasNext();) {
-            String signatureField = i.next();
+        Map<String, BodyHasherImpl> bodyHashJobs = new HashMap<>();
+        Hashtable<String, FailException> signatureExceptions = new Hashtable<>();
+        for (String signatureField : fields) {
             try {
                 int pos = signatureField.indexOf(':');
                 if (pos > 0) {
-                    String v = signatureField.substring(pos + 1, signatureField
-                            .length());
-                    SignatureRecord signatureRecord;
+                    String v = signatureField.substring(pos + 1);
+                    SignatureRecord signatureRecord = null;
                     try {
                         signatureRecord = newSignatureRecord(v);
-                        // validate
-                        signatureRecord.validate();
                     } catch (IllegalStateException e) {
-                        throw new PermFailException("Invalid signature record: "+e.getMessage(), e);
+                        throw new PermFailException("Invalid signature record: " + e.getMessage(), signatureRecord, e);
                     }
 
                     // Specification say we MAY refuse to verify the signature.
                     if (signatureRecord.getSignatureTimestamp() != null) {
-                        long signedTime = signatureRecord.getSignatureTimestamp().longValue();
-                        long elapsed = (System.currentTimeMillis()/1000 - signedTime);
-                        if (elapsed < -3600*24*365*3) {
+                        long signedTime = signatureRecord.getSignatureTimestamp();
+                        long elapsed = (System.currentTimeMillis() / 1000 - signedTime);
+                        if (elapsed < -3600 * 24 * 365 * 3) {
                             throw new PermFailException("Signature date is more than "
-                                    + -elapsed/(3600*24*365) + " years in the future.");
-                        } else if (elapsed < -3600*24*30*3) {
+                                    + -elapsed / (3600 * 24 * 365) + " years in the future.", signatureRecord);
+                        } else if (elapsed < -3600 * 24 * 30 * 3) {
                             throw new PermFailException("Signature date is more than "
-                                    + -elapsed/(3600*24*30) + " months in the future.");
-                        } else if (elapsed < -3600*24*3) {
+                                    + -elapsed / (3600 * 24 * 30) + " months in the future.", signatureRecord);
+                        } else if (elapsed < -3600 * 24 * 3) {
                             throw new PermFailException("Signature date is more than "
-                                        + -elapsed/(3600*24) + " days in the future.");
-                        } else if (elapsed < -3600*3) {
+                                    + -elapsed / (3600 * 24) + " days in the future.", signatureRecord);
+                        } else if (elapsed < -3600 * 3) {
                             throw new PermFailException("Signature date is more than "
-                                    + -elapsed/3600 + " hours in the future.");
-                        } else if (elapsed < -60*3) {
+                                    + -elapsed / 3600 + " hours in the future.", signatureRecord);
+                        } else if (elapsed < -60 * 3) {
                             throw new PermFailException("Signature date is more than "
-                                    + -elapsed/60 + " minutes in the future.");
+                                    + -elapsed / 60 + " minutes in the future.", signatureRecord);
                         } else if (elapsed < 0) {
                             throw new PermFailException("Signature date is "
-                                    + elapsed + " seconds in the future.");
+                                    + elapsed + " seconds in the future.", signatureRecord);
                         }
                     }
 
@@ -308,9 +304,7 @@ public class DKIMVerifier extends DKIMCommon {
                     throw new PermFailException(
                             "unexpected bad signature field");
                 }
-            } catch (TempFailException e) {
-                signatureExceptions.put(signatureField, e);
-            } catch (PermFailException e) {
+            } catch (TempFailException | PermFailException e) {
                 signatureExceptions.put(signatureField, e);
             } catch (RuntimeException e) {
                 signatureExceptions.put(signatureField, new PermFailException(
@@ -318,37 +312,26 @@ public class DKIMVerifier extends DKIMCommon {
             }
         }
 
-        if (bodyHashJobs.isEmpty()) {
-            if (signatureExceptions.size() > 0) {
-                throw prepareException(signatureExceptions);
-            } else {
-                throw new PermFailException("Unexpected condition with "+fields);
-            }
-        }
-
         return new CompoundBodyHasher(bodyHashJobs, signatureExceptions);
     }
 
     /**
      * Verifies all of the DKIM-Signature records declared in the Headers
      * object.
-     * 
-     * @param messageHeaders
-     *                parsed headers
-     * @param bodyInputStream
-     *                input stream for the body.
+     *
+     * @param messageHeaders  parsed headers
+     * @param bodyInputStream input stream for the body.
      * @return a list of verified signature records
-     * @throws IOException
-     * @throws FailException
-     *                 if no signature can be verified
+     * @throws IOException If error occurs handling data
+     * @throws FailException if no signature can be verified
      */
     public List<SignatureRecord> verify(Headers messageHeaders,
-            InputStream bodyInputStream) throws IOException, FailException {
-        
+                                        InputStream bodyInputStream) throws IOException, FailException {
+
         BodyHasher bh = newBodyHasher(messageHeaders);
-        
+
         if (bh == null) return null;
-        
+
         CompoundBodyHasher cbh = validateBodyHasher(bh);
 
         // simultaneous computation of all the hashes.
@@ -356,27 +339,27 @@ public class DKIMVerifier extends DKIMCommon {
 
         return verify(cbh);
     }
-    
+
     /**
-     * Completes the simultaneous verification of multiple 
+     * Completes the simultaneous verification of multiple
      * signatures given the previously prepared compound body hasher where
-     * the user already written the body to the outputstream and closed it. 
-     * 
-     * @param compoundBodyHasher the BodyHasher previously obtained by this class.
+     * the user already written the body to the outputstream and closed it.
+     *
+     * @param bh the BodyHasher previously obtained by this class.
      * @return a list of valid (verified) signatures or null on null input.
      * @throws FailException if no valid signature is found
      */
     public List<SignatureRecord> verify(BodyHasher bh) throws FailException {
         if (bh == null) return null;
         CompoundBodyHasher cbh = validateBodyHasher(bh);
-        
+
         return verify(cbh);
     }
 
     /**
      * Used by public "verify" methods to make sure the input
      * bodyHasher is a CompoundBodyHasher as expected.
-     *  
+     *
      * @param bh the BodyHasher previously obtained by this class.
      * @return a casted CompoundBodyHasher
      * @throws PermFailException if it wasn't a CompoundBodyHasher
@@ -386,97 +369,138 @@ public class DKIMVerifier extends DKIMCommon {
         if (!(bh instanceof CompoundBodyHasher)) {
             throw new PermFailException("Unexpected BodyHasher type: this is not generated by DKIMVerifier!");
         }
-        
-        CompoundBodyHasher cbh = (CompoundBodyHasher) bh;
-        return cbh;
+
+        return (CompoundBodyHasher) bh;
     }
 
     /**
-     * Internal method to complete the simultaneous verification of multiple 
+     * Internal method to complete the simultaneous verification of multiple
      * signatures given the previously prepared compound body hasher where
-     * the user already written the body to the outputstream and closed it. 
-     * 
+     * the user already written the body to the outputstream and closed it.
+     *
      * @param compoundBodyHasher the BodyHasher previously obtained by this class.
      * @return a list of valid (verified) signatures
      * @throws FailException if no valid signature is found
      */
     private List<SignatureRecord> verify(CompoundBodyHasher compoundBodyHasher)
             throws FailException {
-        List<SignatureRecord> verifiedSignatures = new LinkedList<SignatureRecord>();
-        for (Iterator<BodyHasherImpl> i = compoundBodyHasher.getBodyHashJobs().values().iterator(); i.hasNext();) {
-            BodyHasherImpl bhj = i.next();
-
+        List<SignatureRecord> verifiedSignatures = new LinkedList<>();
+        for (BodyHasherImpl bhj : compoundBodyHasher.getBodyHashJobs().values()) {
             byte[] computedHash = bhj.getDigest();
             byte[] expectedBodyHash = bhj.getSignatureRecord().getBodyHash();
 
             if (!Arrays.equals(expectedBodyHash, computedHash)) {
                 compoundBodyHasher.getSignatureExceptions()
                         .put(
-                                "DKIM-Signature:"+bhj.getSignatureRecord().toString(),
+                                "DKIM-Signature:" + bhj.getSignatureRecord().toString(),
                                 new PermFailException(
-                                        "Computed bodyhash is different from the expected one"));
+                                        "Computed bodyhash is different from the expected one", bhj.getSignatureRecord()));
             } else {
                 verifiedSignatures.add(bhj.getSignatureRecord());
             }
         }
 
+        for(SignatureRecord s: verifiedSignatures) {
+            result.add(new Result(s));
+        }
+        result.addAll(resultsFromExceptions(compoundBodyHasher.getSignatureExceptions()));
+
         if (verifiedSignatures.isEmpty()) {
             throw prepareException(compoundBodyHasher.getSignatureExceptions());
         } else {
-            // There is no access to the signatureExceptions when
-            // there is at least one valid signature (JDKIM-14)
-            /*
-            for (Iterator i = signatureExceptions.keySet().iterator(); i
-                    .hasNext();) {
-                String f = (String) i.next();
-                System.out.println("DKIM-Error:"
-                        + ((FailException) signatureExceptions.get(f))
-                                .getMessage() + " FIELD: " + f);
-            }
-            */
-            /*
-            for (Iterator i = verifiedSignatures.iterator(); i.hasNext();) {
-                SignatureRecord sr = (SignatureRecord) i.next();
-                System.out.println("DKIM-Pass:" + sr);
-            }
-            */
             return verifiedSignatures;
         }
     }
 
+    /**
+     * Returns the results of all signature checks, success and fail.
+     * If a message doesn't have a DKIM-Signature header the result list will
+     * be empty. Use {@link Result#getResultType} or {@link Result#isSuccess}
+     * to get verification result of each DKIM-Signature check. Call {@link #resetResults}
+     * if the same instance of {@code DKIMVerifier} is reused for a new {@code verify} call.
+     *
+     * @return List of {@link Result} object.
+     */
+    public List<Result> getResults() {
+        return result;
+    }
+
+    /**
+     * Returns {@code true} if at least one signature is successfully verified.
+     * A message pass the DKIM check when at least one signature is valid.
+     * This method is DMARC compliant, and should be called after {@code DKIMVerifier.verify}.
+     * Use {@link #getResults} to get result details.
+     *
+     * @return {@code true} if a valid signature is found,
+     * {@code false} if there's no valid signature.
+     * @see #getResults
+     */
+    public boolean hasAnyValidSignature() {
+        return result.stream().anyMatch(Result::isSuccess);
+    }
+
+    /**
+     * Clears results list for the DKIMVerifier instance
+     */
+    public void resetResults() {
+        result.clear();
+    }
+
+    private List<Result> resultsFromExceptions(Map<String, FailException> exceptions) {
+        List<Result> results = new ArrayList<>();
+        for (Map.Entry<String, FailException> e : exceptions.entrySet()) {
+            SignatureRecord rec = e.getValue().getRelatedRecord();
+            if (rec == null) {
+                rec = new SignatureRecordTemplate("v=1; d=invalid; h=from; s=invalid; b=invalidsig");
+            }
+
+            Result.Type resultType = Result.Type.NONE;
+            if (e.getValue() instanceof TempFailException) {
+                resultType = Result.Type.TEMPERROR;
+            } else if (e.getValue() instanceof PermFailException) {
+                if (e.getValue().getRelatedRecord() == null) {
+                    //FailException without the SignatureRecord
+                    resultType = Result.Type.PERMERROR;
+                } else {
+                    resultType = Result.Type.FAIL;
+                }
+            }
+            results.add(new Result(e.getValue().getMessage(), e.getKey() != null ? e.getKey() : "", rec, resultType));
+        }
+
+        return results;
+    }
+
     /**
      * Given a map of exceptions prepares a human readable exception.
      * This simply return the exception if it is only one, otherwise returns
      * a cumulative exception
-     * 
+     *
      * @param signatureExceptions input exceptions
-     * @return the resulting "compact" exception 
+     * @return the resulting "compact" exception
      */
     private FailException prepareException(Map<String, FailException> signatureExceptions) {
         if (signatureExceptions.size() == 1) {
             return signatureExceptions.values().iterator()
                     .next();
         } else {
-            // TODO loops signatureExceptions to give a more complete
-            // response, using nested exception or a compound exception.
-            // System.out.println(signatureExceptions);
-            return new PermFailException("found " + signatureExceptions.size()
+            return new CompositeFailException(signatureExceptions.values(), "found " + signatureExceptions.size()
                     + " invalid signatures");
         }
     }
 
     /**
      * Performs signature verification (excluding the body hash).
-     * 
-     * @param h the headers
-     * @param sign the signature record
+     *
+     * @param h       the headers
+     * @param sign    the signature record
      * @param decoded the expected signature hash
-     * @param key the DKIM public key record
+     * @param key     the DKIM public key record
      * @param headers the list of signed headers
-     * @throws PermFailException
+     * @throws PermFailException If signature or public key validation fails
      */
     private void signatureVerify(Headers h, SignatureRecord sign,
-            byte[] decoded, PublicKeyRecord key, List<CharSequence> headers)
+                                 byte[] decoded, PublicKeyRecord key, List<CharSequence> headers)
             throws PermFailException {
         try {
             Signature signature = Signature.getInstance(sign.getHashMethod()
@@ -486,20 +510,16 @@ public class DKIMVerifier extends DKIMCommon {
             try {
                 publicKey = key.getPublicKey();
             } catch (IllegalStateException e) {
-                throw new PermFailException("Invalid Public Key: "+e.getMessage(), e);
+                throw new PermFailException("Invalid Public Key: " + e.getMessage(), sign, e);
             }
             signature.initVerify(publicKey);
 
             signatureCheck(h, sign, headers, signature);
 
             if (!signature.verify(decoded))
-                throw new PermFailException("Header signature does not verify");
-        } catch (InvalidKeyException e) {
-            throw new PermFailException(e.getMessage(), e);
-        } catch (NoSuchAlgorithmException e) {
-            throw new PermFailException(e.getMessage(), e);
-        } catch (SignatureException e) {
-            throw new PermFailException(e.getMessage(), e);
+                throw new PermFailException("Header signature does not verify", sign);
+        } catch (InvalidKeyException | NoSuchAlgorithmException | SignatureException e) {
+            throw new PermFailException(e.getMessage(), sign, e);
         }
     }
 
